#include "Color.h"

Color::Color(unsigned char r, unsigned char g, unsigned char b, double a)
    : mR(r),
      mG(g),
      mB(b),
      mAlpha(a)
{

}

Color::Color(const String& hexString)
{
    if (hexString.size() < 7)
    {
        return;
    }
    if (hexString[0] != '#')
    {
        return;
    }
    mR = toDecimal(hexString.substr(1, 2));
    mG = toDecimal(hexString.substr(3, 2));
    mB = toDecimal(hexString.substr(5, 2));
}

unsigned char Color::toDecimal(const String& hex) const
{
    return static_cast<unsigned char>(std::stoul("0x" + hex, nullptr, 16));
}

Ptr<Color> Color::Create(unsigned char r, unsigned char g, unsigned char b, double a )
{
    return std::make_unique<Color>(r, g, b, a);
}

std::size_t Color::getSize() const
{
    return getSize(mFormat, mBitDepth);
}

std::size_t Color::getSize(Format format, unsigned bitDepth)
{
    return getNumChannels(format) * bitDepth / 8;
}

std::size_t Color::getNumChannels(Format format)
{
    if (format == Format::GRAYSCALE || format == Format::LUT)
    {
        return 1;
    }
    else
    {
        return 4;
    }
}

uint8_t Color::getByte(std::size_t index) const
{
    if (mFormat == Format::GRAYSCALE || mFormat == Format::LUT)
    {
        if (index > 0)
        {
            throw std::range_error("Color index out of range in getByte()");
        }
        return mValue;
    }
    else if (mFormat == Format::ARGB)
    {
        switch(index)
        {
            case 0:
                return mAlpha * 255;
            case 1:
                return mR;
            case 2:
                return mG;
            case 3:
                return mB;
            default:
                throw std::range_error("Color index out of range in getByte()");
        }
    }
    else
    {
        switch(index)
        {
            case 0:
                return mR;
            case 1:
                return mG;
            case 2:
                return mB;
            case 3:
                return mAlpha * 255;
            default:
                throw std::range_error("Color index out of range in getByte()");
        }      
    }
}

void Color::setAlpha(float alpha)
{
    mAlpha = static_cast<double>(alpha);
}

Ptr<Color> Color::Create(const Color& color)
{
    return std::make_unique<Color>(color);
}

unsigned char Color::getR() const
{
    return mR;
}

unsigned char Color::getG() const
{
    return mG;
}

unsigned char Color::getB() const
{
    return mB;
}

double Color::getAlpha() const
{
    return mAlpha;
}

Vector<double> Color::getAsVectorDouble() const
{
    return { mR / 255.0, mG / 255.0, mB / 255.0, mAlpha };
}

uint32_t Color::getAsUInt32() const
{
    return static_cast<uint32_t>(mB + (mG << 8) + (mR << 16));
}

String Color::toString() const
{
    return std::to_string(static_cast<int>(mR)) + "," + std::to_string(static_cast<int>(mG)) + "," + std::to_string(static_cast<int>(mB));
}