#include "StringUtils.h"

#include <locale>
#include <algorithm>
#include <sstream>
#include <iomanip>

bool StringUtils::isAlphaNumeric(char c)
{
    return std::isalnum(c);
}

bool StringUtils::isSpace(char c)
{
    return std::isspace(c);
}

bool StringUtils::isAlphabetical(char c)
{
    return std::isalpha(c);
}

Vector<unsigned char> StringUtils::toBytes(const String& input)
{
    return {input.begin(), input.end()};
}

String StringUtils::toString(const Vector<unsigned char>& bytes)
{
    return {bytes.begin(), bytes.end()};
}

Vector<String> StringUtils::toLines(const String& input)
{
    auto result = Vector<String>{};
    auto ss = Stringstream{input};
    for (String line; std::getline(ss, line, '\n');)
    {
        result.push_back(line);
    }
    return result;
}

bool StringUtils::isWhitespaceOnly(const String& input)
{
    if (input.empty())
    {
        return true;
    }
    else
    {
        return std::all_of(input.cbegin(), input.cend(), [](char c){ return std::isspace(c); });
    }
}

std::size_t StringUtils::countFirstConsecutiveHits(const String& input, char c)
{
    auto found_id = input.find(c);
    if(found_id == String::npos)
    {
        return 0;
    }
    else
    {
        std::size_t count = 1;
        for(std::size_t idx=found_id+1; idx<input.size(); idx++)
        {
            if(input[idx] == c)
            {
                count++;
            }
            else
            {
                return count;
            }
        }
        return count;
    }
}

String StringUtils::stripSurroundingWhitepsace(const String& input)
{
    if (input.empty())
    {
        return {};
    }

    std::size_t first_nonspace = 0;
    std::size_t last_nonspace = input.size() - 1;
    for (std::size_t idx = 0; idx < input.size(); idx++)
    {
        if (!std::isspace(input[idx]))
        {
            first_nonspace = idx;
            break;
        }
    }

    if (first_nonspace == last_nonspace)
    {
        return {};
    }

    for (std::size_t idx = last_nonspace; idx > 0; idx--)
    {
        if (!std::isspace(input[idx]))
        {
            last_nonspace = idx;
            break;
        }
    }
    return input.substr(first_nonspace, last_nonspace-first_nonspace + 1);
}

Vector<String> StringUtils::split(const String& input)
{
    Vector<String> substrings;
    String working_string;
    std::locale loc;
    bool last_was_nonspace{ false };
    for (auto c : input)
    {
        if (std::isspace(c, loc))
        {
            if (last_was_nonspace)
            {
                substrings.push_back(working_string);
                working_string = "";
                last_was_nonspace = false;
            }
        }
        else
        {
            last_was_nonspace = true;
            working_string += c;
        }
    }
    if (!working_string.empty())
    {
        substrings.push_back(working_string);
    }
    return substrings;
}

String StringUtils::toLower(const String& s)
{
    String ret;
    std::transform(s.begin(), s.end(), ret.begin(), [](unsigned char c){ return std::tolower(c); });
    return ret;
}

String StringUtils::toPaddedString(unsigned numBytes, unsigned entry)
{
    Stringstream sstr;
    sstr << std::setfill('0') << std::setw(numBytes) << entry;
    return sstr.str();
}

String StringUtils::stripQuotes(const String& input)
{
    if (input.size() < 3)
    {
        return input;
    }
    std::size_t start_index = 0;
    std::size_t end_index = input.size()-1;
    if (input[start_index] == '"')
    {
        start_index = 1;
    }
    if (input[end_index] == '"')
    {
        end_index = end_index - 1;
    }
    return input.substr(start_index, end_index - start_index + 1);
}

String StringUtils::removeUpTo(const String& input, const String& prefix)
{
    std::size_t found = input.find(prefix);
    if (found != String::npos)
    {
        return input.substr(found + prefix.size(), input.size()-found);
    }
    else
    {
        return input;
    }
}

bool StringUtils::startsWith(const String& input, const String& prefix, bool ignoreWhitespace)
{
    if(ignoreWhitespace)
    {
        const auto loc = input.find(prefix);
        if (loc == String::npos)
        {
            return false;
        }
        else
        {
            return isWhitespaceOnly(input.substr(0, loc));
        }
    }
    else
    {
        return input.find(prefix) == 0;
    }
}
