#include "Image.h"

#include "Color.h"
#include "Grid.h"

#include <cassert>

#ifdef _WIN32
#include "Win32WicImage.h"
#endif

Image::Image(unsigned width, unsigned height, Color::Format format)
    :   mWidth(width),
        mHeight(height),
        mFormat(format)
{
    initialize();
}

Image::~Image()
{

}

Ptr<Image> Image::Create(unsigned width, unsigned height, Color::Format format)
{
    return std::make_unique<Image>(width, height, format);
}

void Image::initialize()
{
    Bounds bounds(0.0, mWidth, 0.0, mHeight);
    const auto pixel_size = Color::getSize(mFormat, mBitDepth);
    mGrid = std::make_unique<SerializeableGrid<Pixel> >(pixel_size, bounds, mWidth, mHeight);
}

void Image::setPixelValues(const Indices& indices, const Vector<Color>& colors)
{
    if (!mGrid)
    {
        initialize();
    }

    assert(indices.size() == colors.size());

    for (std::size_t idx=0; idx< indices.size(); idx++)
    {
        const auto id = indices[idx];
        mGrid->getItem(id.first, id.second).setColor(colors[idx]);
    }
}

unsigned Image::getBytesPerRow() const
{
    return mWidth * Color::getSize(mFormat, mBitDepth);
}

unsigned Image::getWidth() const
{
    return mWidth;
}

unsigned Image::getHeight() const
{
    return mHeight;
}

unsigned Image::getBitDepth() const
{
    return mBitDepth;
}

PlatformImage* Image::getPlatformImage()
{
    if (!mPlatformImage)
    {
#ifdef _WIN32
        mPlatformImage = std::make_unique<Win32WicImage>(this);
#endif
    }
    return mPlatformImage.get();
}

SerializeableGrid<Pixel>* Image::getGrid() const
{
    return mGrid.get();
}

uint8_t Image::getByte(std::size_t index) const
{
    return mGrid->getByte(index);
}

void Image::toBuffer(uint8_t* buffer, std::size_t bufferMaxSize)
{
    mGrid->toBuffer(buffer, bufferMaxSize);
}

Color::Format Image::getFormat() const
{
    return mFormat;
}

unsigned Image::getNumChannels() const
{
    return Color::getNumChannels(mFormat);
}

void Image::setWidth(unsigned width)
{
    mWidth = width;
}

void Image::setHeight(unsigned height)
{
    mHeight = height;
}

void Image::setBitDepth(unsigned bitDepth)
{
    mBitDepth = bitDepth;
}

void Image::setFormat(Color::Format format)
{
    mFormat = format;
}

